# DirTree.tcl --
#
#	Implements directory tree for Unix file systems
#
#       What the indicators mean:
#
#	(+): There are some subdirectories in this directory which are not
#	     currently visible.
#	(-): This directory has some subdirectories and they are all visible
#
#      none: The dir has no subdirectori(es).
#
# Copyright (c) 1996, Expert Interface Technologies
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

tixWidgetClass tixDirTree {
    -classname TixDirTree
    -superclass tixVTree
    -method {
	activate chdir refresh
    }
    -flag {
	-browsecmd -command -directory -disablecallback -showhidden -value
    }
    -configspec {
	{-browsecmd browseCmd BrowseCmd {}}
	{-command command Command {}}
	{-directory directory Directory {}}
	{-disablecallback disableCallback DisableCallback 0 tixVerifyBoolean}
	{-showhidden showHidden ShowHidden 0 tixVerifyBoolean}
    }
    -alias {
	{-value	-directory}
    }
    -default {
	{.scrollbar			auto}
	{*Scrollbar.background          #d9d9d9}
	{*Scrollbar.relief              sunken}
	{*Scrollbar.takeFocus           0}
	{*Scrollbar.troughColor         #c3c3c3}
	{*Scrollbar.width               15}
	{*borderWidth                   1}
	{*hlist.indicator               1}
	{*hlist.background              #c3c3c3}
	{*hlist.drawBranch              1}
	{*hlist.height                  10}
	{*hlist.highlightBackground      #d9d9d9}
	{*hlist.indent                  20}
	{*hlist.itemType                imagetext}
	{*hlist.padX                    3}
	{*hlist.padY                    0}
	{*hlist.relief                  sunken}
	{*hlist.takeFocus               1}
	{*hlist.wideSelection           0}
	{*hlist.width                   20}
    }
}

proc tixDirTree::InitWidgetRec {w} {
    upvar #0 $w data

    tixChainMethod $w InitWidgetRec

    if {$data(-directory) == {}} {
	global env
	if {[info exists env(PWD)]} {
	    set data(-directory) $env(PWD)
	} else {
	    set data(-directory) [pwd]
	}
    }
}

proc tixDirTree::ConstructWidget {w} {
    upvar #0 $w data

    tixChainMethod $w ConstructWidget
    tixDoWhenMapped $w "tixDirTree::StartUp $w"

    $data(w:hlist) config \
	-separator "/" \
	-selectmode "single" -drawbranch 1
}

proc tixDirTree::SetBindings {w} {
    upvar #0 $w data

    tixChainMethod $w SetBindings

# %% do I still need this?
#   bind $data(w:hlist) <3> "tixDirTree::DeleteSib $w %x %y"
}

proc tixDirTree::DeleteSib {w x y} {
    upvar #0 $w data

    set ent [$data(w:hlist) nearest $y]

    if {$ent != ""} {
	$data(w:hlist) anchor set $ent

	for {set e $ent} {$e != "/"} {set e [$data(w:hlist) info parent $e]} {
	    $data(w:hlist) delete siblings $e
	}
	tixDirTree::Browse $w $ent
    }
}

# %% This functions needs to be optimized
#
#
proc tixDirTree::HasSubDir {w dir} {
    upvar #0 $w data

    if $data(-showhidden) {
	set pattern "$dir/* $dir/.*"
    } else {
	set pattern $dir/*
    }

    if [catch {set names [eval glob -nocomplain $pattern]}] {
	# Cannot read directory
	# %% todo: show directory permission denied
	return 0
    }

    # Get rid of . and ..
    #
    foreach f $names {
	if ![string compare [file tail $f] "."] {
	    continue
	}
	if ![string compare [file tail $f] ".."] {
	    continue
	}
	if [file isdirectory $f] {
	    return 1
	}
    }
    return 0
}

proc tixDirTree::AddToList {w dir name image} {
    upvar #0 $w data

    $data(w:hlist) add $dir -text $name -image $image

    if [tixDirTree::HasSubDir $w $dir] {
	tixVTree::SetMode $w $dir open
    }
}

# Add $dir and all ancestors of $dir into the HList widget
#
#
proc tixDirTree::AddDir {w dir} {
    upvar #0 $w data

    if [$data(w:hlist) info exists $dir] {
	return
    }

    if {$dir != "/"} {
	set parentDir [file dirname $dir]

	# This is a bug fix for Tcl 7.5 a2
	#
	if {![string compare $parentDir ":"]} {
	    set parentDir /
	}
	if {![string compare $parentDir "."]} {
	    set parentDir /
	}

	if {![$data(w:hlist) info exists $parentDir]} {
	    tixDirTree::AddDir $w $parentDir
	}
	tixDirTree::AddToList $w $dir [file tail $dir] \
	    [tix getimage folder]
    } else {
	$data(w:hlist) add $dir -text / \
	    -image [tix getimage folder]
	tixVTree::SetMode $w $dir open
    }
}


# Add all the sub directories of $dir into the HList widget
#
#
proc tixDirTree::ListDirs {w dir} {
    upvar #0 $w data

    set appPWD [pwd]

    if [catch {cd $dir} err] {
	# The user has entered an invalid directory
	# %% todo: prompt error, go back to last succeed directory
	cd $appPWD
	return
    }

    if $data(-showhidden) {
	if [catch {set names [lsort [glob -nocomplain * .*]]} err] {
	    # Cannot read directory
	    # %% todo: show directory permission denied
	    cd $appPWD
	    return
	}
    } else {
	if [catch {set names [lsort [glob -nocomplain *]]} err] {
	    cd $appPWD
	    return
	}
    }

    foreach fname $names {
	if {![string compare . $fname] || ![string compare .. $fname]} {
	    continue
	}
	if {$dir == "/"} {
	    set subdir /$fname
	} else {
	    set subdir $dir/$fname
	}
	if [file isdirectory $subdir] {
	    if {![$data(w:hlist) info exists $subdir]} {
		tixDirTree::AddToList $w $subdir $fname \
		    [tix getimage folder]
	    }
	}
    }

    cd $appPWD
}

proc tixDirTree::LoadDir {w dir {mode toggle}} {
    if {![winfo exists $w]} {
	return
    }

    upvar #0 $w data

    # Add the directory and set it to the active directory
    #
    set data(oldDir) $dir
    tixDirTree::AddDir $w $dir
    $data(w:hlist) entryconfig $dir -image [tix getimage act_folder]

    if {$mode == "toggle"} {
	if {[$data(w:hlist) info children $dir] == {}} {
	    set mode expand
	} else {
	    set mode flatten
	}
    }

    if {$mode == "expand"} {
	tixDirTree::ListDirs $w $dir
	if {[$data(w:hlist) info children $dir] == {}} {
	    tixVTree::SetMode $w $dir none
	} else {
	    tixVTree::SetMode $w $dir close
	}
    } else {
	$data(w:hlist) delete offsprings $dir
	tixVTree::SetMode $w $dir open
    }
}

proc tixDirTree::ToggleDir {w value mode} {
    upvar #0 $w data

    tixDirTree::LoadDir $w $value $mode
    tixDirTree::CallCommand $w
}

proc tixDirTree::CallCommand {w} {
    upvar #0 $w data

    if {$data(-command) != {} && !$data(-disablecallback)} {
	set bind(specs) {%V}
	set bind(%V)    $data(-directory)

	tixEvalCmdBinding $w $data(-command) bind $data(-directory)
    }
}

proc tixDirTree::CallBrowseCmd {w ent} {
    upvar #0 $w data

    if {$data(-browsecmd) != {} && !$data(-disablecallback)} {
	set bind(specs) {%V}
	set bind(%V)    $data(-directory)

	tixEvalCmdBinding $w $data(-browsecmd) bind $data(-directory)
    }
}

proc tixDirTree::StartUp {w} {
    if {![winfo exists $w]} {
	return
    }

    upvar #0 $w data

    tixDirTree::LoadDir $w $data(-directory)
}

proc tixDirTree::ChangeDir {w value {forced 0}} {
    upvar #0 $w data

    if {!$forced && $data(-directory) == $value} {
	return
    }

    if {!$forced && [$data(w:hlist) info exists $value]} {
	# Set the old directory to "non active"
	#
	if [$data(w:hlist) info exists $data(-directory)] {
	    $data(w:hlist) entryconfig $data(-directory) \
		-image [tix getimage folder]
	}

	$data(w:hlist) entryconfig $value  \
		-image [tix getimage act_folder]

    } else {
	if {$forced} {
	    if {[$data(w:hlist) info children $value] == {}} {
		set mode flatten
	    } else {
		set mode expand
	    }
	} else {
	    set mode toggle
	}
	tixDirTree::LoadDir $w $value $mode
	tixDirTree::CallCommand $w
    }
    set data(-directory) $value
}


#----------------------------------------------------------------------
#
# Virtual Methods
#
#----------------------------------------------------------------------
proc tixDirTree::OpenCmd {w ent} {
    tixDirTree::ToggleDir $w $ent expand
    tixDirTree::ChangeDir $w $ent
    tixDirTree::CallBrowseCmd $w $ent
}

proc tixDirTree::CloseCmd {w ent} {
    tixDirTree::ToggleDir $w $ent flatten
    tixDirTree::ChangeDir $w $ent
    tixDirTree::CallBrowseCmd $w $ent
}

proc tixDirTree::Command {w B} {
    upvar #0 $w data
    upvar $B bind

    set ent [tixEvent flag V]
    tixChainMethod $w Command $B

    if {$data(-command) != {}} {
	tixEvalCmdBinding $w $data(-command) bind $ent
    }
}

# This is a virtual method
#
proc tixDirTree::BrowseCmd {w B} {
    upvar #0 $w data
    upvar $B bind

    set ent [tixEvent flag V]

#    if {[$data(w:hlist) indicator exist $ent] && 
#	[$data(w:hlist) info children $ent] == {}} {
#	
#	tixVTree::Activate $w $ent open
#   }

    tixDirTree::ChangeDir $w $ent

    if {$data(-browsecmd) != {}} {
	tixEvalCmdBinding $w $data(-browsecmd) bind $ent
    }
}

#----------------------------------------------------------------------
#
# Public Methods
#
#----------------------------------------------------------------------
proc tixDirTree::chdir {w value} {
    tixDirTree::ChangeDir $w $value
}

proc tixDirTree::refresh {w {dir {}}} {
    upvar #0 $w data

    if {$dir == {}} {
	set dir $data(-directory)
    }

    tixDirTree::ChangeDir $w $dir 1

    foreach sub [$data(w:hlist) info children $dir] {
	if {![file exists $sub]} {
	    $data(w:hlist) delete entry $sub
	}
    }
}

proc tixDirTree::config-directory {w value} {
    tixDirTree::ChangeDir $w $value
}
tSingle <Right> {
	tixHListSingle::LeftRight %W right
    }
    tixBind TixHListSingle <Prior> {
	%W yview scroll -1 pages
    }
    tixBind TixHListSingle <Next> {
	%W yview scroll 1 pages
    }
    tixBind TixHListSingle <Return> {
	tixHListSingle::Keyboard-Activate %W 
    }
    tixBind TixHListSingle <space> {
	tixHListSingle::Keyboard-Browse %W 
    }
}



#----------------------------------------------------------------------
#
#
#			 Mouse bindings
#
#
#----------------------------------------------------------------------


proc tixHListSingle::Button-1 {w x y} {
#    if {[$w cget -state] == "disabled"} {
#	return
#    }

    if [$w cget -takefocus] {
	focus $w
    }

    case [tixHListSingle::GetState $w] {
	{0} {
	    # %% Check if not on title
	    tixHListSingle::GoState 1 $w $x $y

	}
    }
}

proc tixHListSingle::DragTimer {w ent} {
    case [tixHListSingle::GetState $w] {
	{1} {
	    # fire up
	}
    }
}


proc tixHListSingle::ButtonRelease-1 {w x y} {
    case [tixHListSingle::GetState $w] {
	{5 16} {
	    tixHListSingle::GoState 6 $w $x $y
	}
	{15} {
	    tixHListSingle::GoState 17 $w $x $y
	}
	{10 11} {
	    tixHListSingle::GoState 18 $w
	}
	{13 20} {
	    tixHListSingle::GoState 14 $w $x $y
	}
	{21} {
	    tixHListSingle::GoState 22 $w
	}
    }
}

proc tixHListSingle::Double-1 {w x y} {
    case [tixHListSingle::GetState $w] {
	{0} {
	    tixHListSingle::GoState 23 $w $x $y
	}
    }
}

proc tixHListSingle::B1-Motion {w x y} {
    case [tixHListSingle::GetState $w] {
	{1} {
	    tixHListSingle::GoState 5 $w $x $y 
	}
	{5 16} {
	    tixHListSingle::GoState 5 $w $x $y 
	}
	{13 20 21} {
	    tixHListSingle::GoState 20 $w $x $y 
	}
    }
}

proc tixHListSingle::B1-Leave {w} {
    case [tixHListSingle::GetState $w] {
	{5} {
	    tixHListSingle::GoState 10 $w
	}
    }
}

proc tixHListSingle::B1-Enter {w x y} {
    case [tixHListSingle::GetState $w] {
	{10 11} {
	    tixHListSingle::GoState 12 $w $x $y
	}
    }
}

proc tixHListSingle::AutoScan {w} {
    case [tixHListSingle::GetState $w] {
	{10 11} {
	    tixHListSingle::GoState 11 $w
	}
    }
}

#----------------------------------------------------------------------
#
#
#			 Key bindings
#
#
#----------------------------------------------------------------------
proc tixHListSingle::Keyboard-Activate {w} {
    if {[tixHListSingle::GetState $w] != 0} {
	return
    }
    set ent [$w info anchor]

    if {$ent == ""} {
	return
    }

    if {[$w cget -selectmode] == "single"} {
	$w select clear
	$w select set $ent
    }

    set command [$w cget -command]
    if {$command != {}} {
	set bind(specs) {%V}
	set bind(%V)    $ent

	tixEvalCmdBinding $w $command bind $ent
    }
}

proc tixHListSingle::Keyboard-Browse {w} {
    if {[tixHListSingle::GetState $w] != 0} {
	return
    }
    set ent [$w info anchor]

    if {$ent == ""} {
	return
    }

    if {[$w cget -selectmode] == "single"} {
	$w select clear
	$w select set $ent
    }

    tixHListSingle::Browse $w $ent
}

proc tixHListSingle::LeftRight {w spec} {
    if {[tixHListSingle::GetState $w] != 0} {
	return
    }

    set anchor [$w info anchor]
    if {$anchor == ""} {
	set anchor [lindex [$w info children] 0]
    }

    set ent $anchor
    while {1} {
	set e $ent
	if {$spec == "left"} {
	    set ent [$w info parent $e]

	    if {$ent == {} || [$w entrycget $ent -state] == "disabled"} {
		set ent [$w info prev $e]
	    }
	} else {
	    set ent [lindex [$w info children $e] 0]

	    if {$ent == {} || [$w entrycget $ent -state] == "disabled"} {
		set ent [$w info next $e]
	    }
	}

	if {$ent == {}} {
	    break
	}
 	if {[$w entrycget $ent -state] == "disabled"} {
	    continue
	}
 	if [$w info hidden $ent] {
	    continue
	}
	break
    }

    if {$ent == {}} {
       return
    }

    $w anchor set $ent
    $w see $ent

    if {[$w cget -selectmode] != "single"} {
	$w select clear
	$w selection set $ent

	tixHListSingle::Browse $w $ent
    }
}

proc tixHListSingle::UpDown {w spec} {
    if {[tixHListSingle::GetState $w] != 0} {
	return
    }
    set anchor [$w info anchor]

    if {$anchor == ""} {
	set anchor [lindex [$w info children] 0]

	if {$anchor == {}} {
	    return
	}

	if {[$w entrycget $anchor -state] != "disabled"} {
	    # That's a good anchor
	    set done 1
	} else {
	    # We search for the first non-disabled entry (downward)
	    set spec next
	}
    }

    set ent $anchor
    set done 0
    # Find the prev/next non-disabled entry
    #
    while {!$done} {
	set ent [$w info $spec $ent]
	if {$ent == {}} {
	    break
	}
	if {[$w entrycget $ent -state] == "disabled"} {
	    continue
	}
	if [$w info hidden $ent] {
	    continue
	}
	break
    }

    if {$ent == {}} {
	return
    } else {
	$w anchor set $ent
	$w see $ent

	if {[$w cget -selectmode] != "single"} {
	    $w select clear
	    $w selection set $ent

	    tixHListSingle::Browse $w $ent
	}
    }
}

proc tixHListSingle::GetNearest {w y} {
    set ent [$w nearest $y]

    if {$ent != {}} {
	if {[$w entrycget $ent -state] == "disabled"} {
	    return {}
	}
    }

    return $ent
}

#----------------------------------------------------------------------
#
#			STATE MANIPULATION
#
#
#----------------------------------------------------------------------
proc tixHListSingle::GetState {w} {
    global $w:priv:state

    if {![info exists $w:priv:state]} {
	set $w:priv:state 0
    }
    return [set $w:priv:state]
}

proc tixHListSingle::SetState {w n} {
    global $w:priv:state

    set $w:priv:state $n
}

proc tixHListSingle::GoState {n w args} {

#    puts "going from [tixHListSingle::GetState $w] --> $n"

    tixHListSingle::SetState $w $n
    eval tixHListSingle::GoState-$n $w $args
}

#----------------------------------------------------------------------
#			States
#----------------------------------------------------------------------
proc tixHListSingle::GoState-0 {w} {
}
proc tixHListSingle::GoState-1 {w x y} {
    set oldEnt [$w info anchor]
    set ent [tixHListSingle::SetAnchor $w $x $y]

    # %% add if is in dragdrop mode
    set dragdrop 0

    if {$ent == {}} {
	tixHListSingle::GoState 0 $w
    }

    set info [$w info item $x $y]
    if {[lindex $info 1] == "indicator"} {
	tixHListSingle::GoState 13 $w $ent $oldEnt
    } elseif {$dragdrop} {
	tixHListSingle::GoState 15 $w
    } else {
	tixHListSingle::GoState 16 $w $ent
    }
}

proc tixHListSingle::GoState-5 {w x y} {
    set oldEnt [$w info anchor]

    set ent [tixHListSingle::SetAnchor $w $x $y]

    if {$ent == {} || $oldEnt == $ent} {
	return
    }

    if {[$w cget -selectmode] != "single"} {
	tixHListSingle::Select $w $ent
	tixHListSingle::Browse $w $ent
    }
}

proc tixHListSingle::GoState-6 {w x y} {
    set ent [tixHListSingle::SetAnchor $w $x $y]

    if {$ent == {}} {
	tixHListSingle::GoState 0 $w
	return
    }
    tixHListSingle::Select $w $ent
    tixHListSingle::Browse $w $ent

    tixHListSingle::GoState 0 $w
}

proc tixHListSingle::GoState-10 {w} {
    tixHListSingle::StartScan $w
}

proc tixHListSingle::GoState-11 {w} {
    global tkPriv

    tixHListSingle::DoScan $w

    set oldEnt [$w info anchor]
    set ent [tixHListSingle::Set